/*
 * blueMail offline mail reader
 * packet list

 Copyright (c) 1996 Kolossvary Tamas <thomas@vma.bme.hu>
 Copyright (c) 1997 John Zero <john@graphisoft.hu>
 Copyright (c) 1998 William McBrine <wmcbrine@clark.net>
 Copyright (c) 2003 Ingo Brueckl <ib@wupperonline.de>

 Distributed under the GNU General Public License.
 For details, see the file COPYING in the parent directory. */


#include <ctype.h>
#include <string.h>
#include "interfac.h"
#include "../common/auxil.h"
#include "../common/error.h"
#include "../common/mysystem.h"


PacketListWindow::PacketListWindow ()
{
  noOfListed = 0;
  packetList = NULL;
  listed = NULL;
  filter = NULL;
  packetStat = NULL;

  const char *spb = bm.resourceObject->get(SortPacketsBy);
  namesort = (spb ? toupper(*spb) == 'N' : true);
}


PacketListWindow::~PacketListWindow ()
{
  delete packetStat;
  delete[] filter;
  DestroyChain();
}


void PacketListWindow::MakeActive ()
{
  list_max_y = LINES - 18;     // 9 lines from top, 3 lines border and header,
                               // 6 lines to bottom
  list_max_x = 62;

  list = new Window(list_max_y + 3, list_max_x + 2,
                    9, (COLS - list_max_x - 2) >> 1, C_PLBORDER);
  list->attrib(C_PLHEADER);
  list->putstring(1, 3, "Packet");
  list->putstring(1, 20, "Date");
  list->putstring(1, 42, "Size");
  list->putstring(1, 49, "Unread");
  list->putstring(1, 56, "Total");

  if (bm.resourceObject->isYes(DrawSortMark))
  {
    const char *smark = "( )";

    if (namesort)
    {
      list->putstring(1, 9, smark);
      list->putch(1, 10, ACS_DARROW);
    }
    else
    {
      list->putstring(1, 24, smark);
      list->putch(1, 25, ACS_UARROW);
    }
  }

  // MakeChain() is public should have been called prior to MakeActive()
  if (!listed) MakeChain();

  // list may indeed be empty, because the contents of the list
  // may have changed (new packet name) while a filter is active,
  // but we know that items exist
  if (noOfItems() == 0)
  {
    delete[] filter;
    filter = NULL;

    MakeChain();
  }

  if (filter)
  {
    chtype fmark[2] = {' ' | C_PLBORDER, FILTER_SIGN | C_PLBORDER};
    list->putchnstr(0, list_max_x, fmark, 2);
  }

  list->delay_update();

  relist();
  Draw();
  Select();
}


void PacketListWindow::Delete ()
{
  delete list;
}


void PacketListWindow::Quit ()
{
  delete[] filter;
  filter = NULL;
}


void PacketListWindow::MakeChain ()
{
  int dummy;

  DestroyChain();

  if (!packetStat) packetStat = new packet_stat(&bm);

  packetList = new file_list(bm.resourceObject->get(PacketDir));

  int noOfFiles = packetList->getNoOfFiles();

  // sort
  packetList->sort(namesort ? FL_SORT_BY_NAME : FL_SORT_BY_NEWEST);

  // filter
  listed = new int[noOfFiles];
  for (int i = 0; i < noOfFiles; i++)
  {
    packetList->gotoFile(i);
    // mark as used
    packetStat->getStat(packetList->getDate(), &dummy, &dummy);

    // initialize listed, so that oneSearch() can be used to check filter
    listed[i] = i;

    if (!filter || (oneSearch(i, filter) == FND_YES))
      listed[noOfListed++] = i;
  }
}


void PacketListWindow::DestroyChain ()
{
  delete packetList;
  packetList = NULL;

  noOfListed = 0;
  delete[] listed;
  listed = NULL;
}


void PacketListWindow::Select ()
{
  packetList->gotoFile(listed[active]);
}


bool PacketListWindow::OpenPacket ()
{
  Select();
  return bm.selectPacket(packetList->getName());
}


int PacketListWindow::noOfItems ()
{
  return noOfListed;
}


char *PacketListWindow::NameDate (char *p)
{
  const char *fname = packetList->getName();
  time_t ftime = packetList->getDate();

  strcpy(p, "          ");
  for (int j = 2; *fname && (*fname != '.' && j < 10); j++) p[j] = *fname++;
  p += 10;

  p += sprintf(p, "%-4.4s    ", fname);
  p += strftime(p, 18, "%d %b %Y %H:%M", localtime(&ftime));

  return p;
}


void PacketListWindow::oneLine (int i)
{
  int unread, total;
  char *p = list->lineBuf;

  packetList->gotoFile(listed[topline + i]);

  p = NameDate(p);
  p += sprintf(p, "%10ld    ", (long) packetList->getSize());

  if (packetStat->getStat(packetList->getDate(), &total, &unread))
    sprintf(p, "%5d %5d  ", unread, total);
  else strcat(p, "             ");

  list->PUTLINE(i, C_PLIST);
}


searchtype PacketListWindow::oneSearch (int l, const char *what)
{
  packetList->gotoFile(listed[l]);
  (void) NameDate(list->lineBuf);
  return (strexcmp(list->lineBuf, what) ? FND_YES : FND_NO);
}


void PacketListWindow::renamePacket ()
{
  int num = 0;
  char msg[24], input[QUERYINPUT + 1];
  const char *fname = packetList->getName();

  sprintf(msg, "Rename %.12s to:", fname);

  if (getNumExt(fname) != -1 || (num = packetList->nextNumExt(fname)) == -1)
    strncpy(input, fname, QUERYINPUT);
  else
  {
    sprintf(input, "%03d", num);
    strncpy(input, ext(fname, input), QUERYINPUT);
  }

  input[QUERYINPUT] = '\0';

  if (interface->QueryBox(msg, input, QUERYINPUT, false) != GOT_ESC &&
      strcmp(fname, input) != 0 && *input != '\0')
  {
    if (packetList->exists(input))
      interface->ErrorWindow("Such a packet already exists");
    else
    {
      Select();

      if (packetList->changeName(input)) MakeChain();
      else interface->ErrorWindow("Rename failed");
    }
  }
}


void PacketListWindow::askFilter ()
{
  char input[QUERYINPUT + 1];

  strncpy(input, (filter ? filter : ""), QUERYINPUT);
  input[QUERYINPUT] = '\0';

  if (interface->QueryBox("Filter on:", input, QUERYINPUT) != GOT_ESC)
  {
    savePos();

    delete[] filter;
    filter = strdupplus(input);
    MakeChain();

    // empty list isn't allowed
    if (noOfListed == 0)
    {
      if (*filter) interface->ErrorWindow("No match");
      delete[] filter;
      filter = NULL;
      MakeChain();
      restorePos();
    }
  }
}


void PacketListWindow::extrakeys (int key)
{
  char fname[MYMAXPATH];

  Select();

  switch (key)
  {
    case 'A':
      interface->changestate(addressbook);
      break;

    case 'K':
    case KEY_DC:
      char msg[43];
      sprintf(msg, "Do you really want to delete %.12s?", packetList->getName());
      if (interface->WarningWindow(msg) == WW_YES)
      {
        mkfname(fname, bm.resourceObject->get(PacketDir), packetList->getName());
        remove(fname);
        checkDel();
        MakeChain();
      }
      if (noOfListed == 0)
      {
        Quit();
        interface->back();
      }
      else interface->update();
      break;

    case 'O':
      namesort = !namesort;
      MakeChain();
      interface->update();
      break;

    case 'R':
      renamePacket();
      interface->update();
      break;

    case '|':
      askFilter();
      interface->update();
      break;

    case KEY_LEFT:
      Move(UP);
      Draw();
      break;

    case KEY_RIGHT:
      Move(DOWN);
      Draw();
      break;
  }
}
